from pydantic import BaseModel, field_serializer, model_validator
from pydantic_extra_types.color import Color
from pydantic_core import SchemaValidator

from typing import List, Literal, Union, Optional


def check_schema(model: BaseModel) -> None:
    schema_validator = SchemaValidator(schema=model.__pydantic_core_schema__)
    schema_validator.validate_python(model.__dict__)


class FHTableEntry(BaseModel):
    value: Union[str, int, float, None]
    background_color: Optional[Color] = None

    @field_serializer("background_color")
    def serialize_color(self, background_color, _info):
        return background_color.as_hex("Long") if background_color is not None else None


class FHTable(BaseModel):
    title: str
    header: List[FHTableEntry] = []
    rows: List[List[FHTableEntry]] = []

    @model_validator(mode="after")
    def rows_wrong_size(self):
        header_len = len(self.header)
        if header_len != 0 and len(self.rows) != 0:
            for row in self.rows:
                if len(row) != header_len:
                    raise ValueError("row lengths do not match header")
        return self

    def add_row(self, row: List[FHTableEntry]):
        self.rows.append(row)
        check_schema(self)

    def add_header(self, header: List[FHTableEntry]):
        self.header = header
        check_schema(self)


class FHFigure(BaseModel):
    title: str
    filename: str
    width: Optional[Union[int, float, str]] = None
    height: Optional[Union[int, float, str]] = None


class FHObject(BaseModel):
    type: Literal["table", "svg_figure"]
    data: Union[FHTable, FHFigure]


class FHJSON(BaseModel):
    data: List[FHObject] = []

    def add_table(self, table):
        self.data.append(FHObject(type="table", data=table))
        check_schema(self)

    def add_svg_figure(self, figure):
        self.data.append(FHObject(type="svg_figure", data=figure))
        check_schema(self)



import pytest
import html5lib
import floods_html as fh

test_obj_1 = fh.FHJSON(
    data=[
        fh.FHObject(
            type="svg_figure",
            data=fh.FHFigure(title="Test Figure", filename="https://website/test.svg", width="30", height=40),
        ),
        fh.FHObject(
            type="table",
            data=fh.FHTable(
                title="Table One",
                header=[
                    fh.FHTableEntry(value=40, background_color="FFFFFF"),
                    fh.FHTableEntry(value="Name", background_color="FFFFFF"),
                ],
                rows=[
                    [fh.FHTableEntry(value="540", background_color="FFFFFF"), fh.FHTableEntry(value=None)],
                    [fh.FHTableEntry(value="540", background_color="FF0000"), fh.FHTableEntry(value=670)],
                ],
            ),
        ),
    ]
)

test_obj_2 = fh.FHJSON(
    data=[
        fh.FHObject(
            type="table",
            data=fh.FHTable(
                title="Table One",
                header=[
                    fh.FHTableEntry(value=40, background_color="FFFFFF"),
                    fh.FHTableEntry(value="Name"),
                    fh.FHTableEntry(value="Naming"),
                ],
                rows=[
                    [
                        fh.FHTableEntry(value="540", background_color="FFFFFF"),
                        fh.FHTableEntry(value=None),
                        fh.FHTableEntry(value=517863),
                    ],
                ],
            ),
        ),
        fh.FHObject(
            type="svg_figure",
            data=fh.FHFigure(title="Test Figure 1", filename="https://website/test.svg", width="30", height=40),
        ),
        fh.FHObject(type="svg_figure", data=fh.FHFigure(title="Test Figure 2", filename="https://website/test.png")),
    ]
)


@pytest.mark.parametrize(
    "json_info, export_type",
    [
        (test_obj_1, "str"),
        (test_obj_2, "str"),
        (test_obj_1, "dict"),
        (test_obj_2, "dict"),
        (test_obj_1, "object"),
        (test_obj_2, "object"),
    ],
)
def test_valid_html_default_construction(json_info, export_type):
    if export_type == "str":
        exported_object = json_info.model_dump_json()
    elif export_type == "dict":
        exported_object = json_info.model_dump()
    elif export_type == "object":
        exported_object = json_info

    html5parser = html5lib.HTMLParser(strict=True)
    for html in fh.json_to_html(exported_object):
        html5parser.parseFragment(html)


@pytest.mark.parametrize("export_type", ["str", "dict", "object"])
def test_valid_html_manual_construction(export_type):
    json_object = fh.FHJSON()
    figure = fh.FHFigure(title="Test Figure", filename="https://website/test.svg", width="30", height=40)
    json_object.add_svg_figure(figure)
    table = fh.FHTable(title="Table One")
    table.add_header([fh.FHTableEntry(value=40, background_color="FFFFFF"), fh.FHTableEntry(value="Name")])
    table.add_row([fh.FHTableEntry(value="540", background_color="FFFFFF"), fh.FHTableEntry(value=None)])
    table.add_row([fh.FHTableEntry(value="540", background_color="FF0000"), fh.FHTableEntry(value=670)])
    json_object.add_table(table)

    if export_type == "str":
        exported_object = json_object.model_dump_json()
    elif export_type == "dict":
        exported_object = json_object.model_dump()
    elif export_type == "object":
        exported_object = json_object

    html5parser = html5lib.HTMLParser(strict=True)
    for html in fh.json_to_html(exported_object):
        html5parser.parseFragment(html)



from floods_html import json_format as jf


def json_to_html(input, **kwargs):
    if type(input) is str:
        pydantic_data_object = jf.FHJSON.model_validate_json(input)
    elif type(input) is dict:
        pydantic_data_object = jf.FHJSON(**input)
    elif isinstance(input, jf.FHJSON):
        pydantic_data_object = input
    else:
        raise ValueError("Invalid input type. Must be either a JSON string, JSON object, or a FHJSON class instance.")
    html_output = []
    for entry in pydantic_data_object.data:
        html_entry = entry_to_html(entry, **kwargs)
        html_output.append(html_entry)
    return html_output


def entry_to_html(entry, **kwargs):
    if entry.type == "table":
        return table_to_html(entry.data, **kwargs)
    elif entry.type == "svg_figure":
        return figure_to_html(entry.data, **kwargs)
    else:
        raise ValueError("Unknown entry type: {}".format(entry.type))


def table_to_html(json, default_colour="FFFFFF", **kwargs):
    html_table_header = ""
    html_table_rows = ""

    table_row_html_template = '<td style="background-color: {colour}">{value}</td>'

    for table_header_entry in json.header:
        html_table_header += table_row_html_template.format(
            colour=table_header_entry.background_color
            if table_header_entry.background_color is not None
            else default_colour,
            value=table_header_entry.value or "",
        )

    for table_row in json.rows:
        html_table_row = ""
        for table_entry in table_row:
            html_table_row += table_row_html_template.format(
                colour=table_entry.background_color if table_entry.background_color is not None else default_colour,
                value=table_entry.value or "",
            )
        html_table_row = "<tr>" + html_table_row + "</tr>"
        html_table_rows += html_table_row

    table_html_template = """
    <h3>{title}</h3>
    <table class="table table-bordered">
        <thead><tr>{header}</tr></thead>
        <tbody>{body}</tbody>
    </table>
    """

    table_html = table_html_template.format(title=json.title, header=html_table_header, body=html_table_rows)

    return table_html


def figure_to_html(json, svg_location="", default_height=100, default_width=100, **kwargs):
    svg_file = svg_location + json.filename

    if svg_file[:4] == "http":
        figure_html_template = """
            <span>
                <h4>{title}</h4>
                <img src={imgname} width={width} height={height} />
            </span>
        """

        figure_html = figure_html_template.format(
            title=json.title,
            imgname=svg_file,
            width=json.width if json.width is not None else default_width,
            height=json.height if json.height is not None else default_height,
        )
    else:
        figure_html_template = """
        <div style="width: {width}px; height: {height}px">
            <span>
                <h4>{title}</h4>
                {svg}
            </span>
        </div>
        """

        svg_contents = open(svg_file, "r").read()
        svg_contents = svg_contents.replace("<svg", '<svg width="100%" height="100%"')

        figure_html = figure_html_template.format(
            title=json.title,
            svg=svg_contents,
            width=json.width if json.width is not None else default_width,
            height=json.height if json.height is not None else default_height,
        )

    return figure_html
